import shutil
import tempfile


class TemporaryDirectory:
    """A context manager for temporary directories"""
    def __init__(self, dir=None):
        self._dir = None
        self._dir = dir
    def __enter__(self):
        self._dir = tempfile.mkdtemp(dir=self._dir)
        return self._dir
    def __exit__(self, exc_type, exc_value, traceback):
        if self._dir:
            shutil.rmtree(self._dir)



import time
from tempfile import TemporaryDirectory
import numpy as np
import spikeinterface as si
import spikeinterface.extractors as se
import spikeinterface.preprocessing as spre
import mountainsort5 as ms5
from mountainsort5.util import create_cached_recording


def test_scheme2():
    recording, sorting_true = se.toy_example(
        duration=40,
        num_channels=8,
        num_units=16,
        sampling_frequency=30000,
        num_segments=1,
        seed=0
    ) # type: ignore
    recording: si.BaseRecording = recording
    sorting_true: si.BaseSorting = sorting_true

    # lazy preprocessing
    recording_filtered = spre.bandpass_filter(recording, freq_min=300, freq_max=6000, dtype=np.float32)
    recording_preprocessed: si.BaseRecording = spre.whiten(recording_filtered)

    with TemporaryDirectory() as tmpdir:
        # cache the recording to a temporary directory for efficient reading
        recording_cached = create_cached_recording(recording_preprocessed, folder=tmpdir)

        # sorting
        print('Starting MountainSort5 (sorting1)')
        timer = time.time()
        result = ms5.sorting_scheme2(
            recording_cached,
            sorting_parameters=ms5.Scheme2SortingParameters(
                phase1_detect_channel_radius=150,
                detect_channel_radius=50,
                max_num_snippets_per_training_batch=3, # for improving test coverage
                snippet_mask_radius=150,
                training_duration_sec=15
            ),
            return_snippet_classifiers=True # for coverage
        )
        assert isinstance(result, tuple)
        sorting1, classifer1 = result

    elapsed_sec = time.time() - timer
    duration_sec = recording.get_total_duration()
    print(f'Elapsed time for sorting: {elapsed_sec:.2f} sec -- x{(duration_sec / elapsed_sec):.2f} speed compared with real time for {recording.get_num_channels()} channels')

    print('Starting MountainSort5 (sorting2)')
    timer = time.time()
    sorting2 = ms5.sorting_scheme2( # noqa
        recording_preprocessed,
        sorting_parameters=ms5.Scheme2SortingParameters(
            phase1_detect_channel_radius=150,
            detect_channel_radius=50,
            training_duration_sec=25,
            training_recording_sampling_mode='uniform'
        )
    )
    elapsed_sec = time.time() - timer
    duration_sec = recording.get_total_duration()
    print(f'Elapsed time for sorting: {elapsed_sec:.2f} sec -- x{(duration_sec / elapsed_sec):.2f} speed compared with real time for {recording.get_num_channels()} channels')

    # commenting out because this step takes a while
    # print('Comparing with truth')
    # comparison = sc.compare_sorter_to_ground_truth(gt_sorting=sorting_true, tested_sorting=sorting)
    # print(comparison.get_performance())

# for coverage
def test_scheme2_single_segment():
    recording, sorting_true = se.toy_example(
        duration=4,
        num_channels=4,
        num_units=4,
        sampling_frequency=30000,
        num_segments=1,
        seed=0
    ) # type: ignore
    recording: si.BaseRecording = recording
    sorting_true: si.BaseSorting = sorting_true

    sorting1 = ms5.sorting_scheme2( # noqa
        recording,
        sorting_parameters=ms5.Scheme2SortingParameters(
            phase1_detect_channel_radius=150,
            detect_channel_radius=50,
            max_num_snippets_per_training_batch=3, # for improving test coverage
            snippet_mask_radius=150,
            training_duration_sec=15,
            classifier_npca=4
        )
    )



from typing import Union, List
import numpy as np
import spikeinterface as si
import spikeinterface.preprocessing as spre
from dataclasses import dataclass
from ..schemes.Scheme1SortingParameters import Scheme1SortingParameters
from ..schemes.sorting_scheme2 import get_time_chunks
from ..core.get_block_recording_for_scheme3 import get_block_recording_for_scheme3
from ..schemes.sorting_scheme1 import sorting_scheme1
from ..schemes.sorting_scheme2 import TimeChunk


@dataclass
class EstimateUnitsParameters:
    block_sorting_parameters: Scheme1SortingParameters = Scheme1SortingParameters()
    avg_num_channels_per_neighborhood: Union[int, None] = 7
    block_duration_sec: float = 300
    max_num_blocks: Union[int, None] = 10

    def check_valid(
        self,
        *,
        M: int,
        N: int,
        sampling_frequency: float,
        channel_locations: np.ndarray,
    ):
        """Internal function for checking validity of parameters"""
        self.block_sorting_parameters.check_valid(
            M=M,
            N=N,
            sampling_frequency=sampling_frequency,
            channel_locations=channel_locations,
        )
        assert self.block_duration_sec > 0


@dataclass
class EstimateUnitsUnit:
    unit_id: Union[int, str]
    num_spikes: int
    peak_channel_id: Union[int, str]
    snr: float

    def to_dict(self):
        return {
            "unit_id": self.unit_id,
            "num_spikes": self.num_spikes,
            "peak_channel_id": self.peak_channel_id,
            "snr": self.snr,
        }


@dataclass
class EstimateUnitsBlock:
    start_time_sec: float
    end_time_sec: float
    units: List[EstimateUnitsUnit]

    def to_dict(self):
        return {
            "start_time_sec": self.start_time_sec,
            "end_time_sec": self.end_time_sec,
            "units": [u.to_dict() for u in self.units],
        }


@dataclass
class EstimateUnitsOutput:
    blocks: List[EstimateUnitsBlock]

    def to_dict(self):
        return {
            "blocks": [b.to_dict() for b in self.blocks],
        }

    def __repr__(self):
        ret = f"EstimateUnitsOutput(blocks={len(self.blocks)})"
        for i, block in enumerate(self.blocks):
            ret += "\n"
            ret += f"  block {i}: {len(block.units)} units"
            for j, unit in enumerate(block.units):
                ret += "\n"
                ret += f"    unit {j}: firing rate {unit.num_spikes / (block.end_time_sec - block.start_time_sec):.2f} Hz, peak channel {unit.peak_channel_id}, snr {unit.snr:.2f}"
        return ret


def estimate_units(
    recording: si.BaseRecording,
    estimate_units_parameters: EstimateUnitsParameters = EstimateUnitsParameters(),
) -> EstimateUnitsOutput:
    ###################################################################
    # Handle multi-segment recordings
    if recording.get_num_segments() > 1:
        raise Exception("Multi-segment recordings not supported for estimate_units")
    ###################################################################

    output = EstimateUnitsOutput(blocks=[])

    M = recording.get_num_channels()
    N = recording.get_num_frames()
    sampling_frequency = recording.sampling_frequency
    channel_locations = recording.get_channel_locations()

    if estimate_units_parameters.avg_num_channels_per_neighborhood is not None:
        radius = _auto_detect_channel_neighborhood_radius(
            channel_locations,
            estimate_units_parameters.avg_num_channels_per_neighborhood,
        )
        if (
            estimate_units_parameters.block_sorting_parameters.snippet_mask_radius
            is None
        ):
            estimate_units_parameters.block_sorting_parameters.snippet_mask_radius = (
                radius
            )
            # raise Exception('TEMP')
        else:
            raise Exception(
                "Cannot specify both avg_num_channels_per_neighborhood and snippet_mask_radius"
            )
        if (
            estimate_units_parameters.block_sorting_parameters.detect_channel_radius
            is None
        ):
            estimate_units_parameters.block_sorting_parameters.detect_channel_radius = (
                radius
            )
        else:
            raise Exception(
                "Cannot specify both avg_num_channels_per_neighborhood and detect_channel_radius"
            )

    if estimate_units_parameters.block_sorting_parameters.skip_alignment is None:
        # cuts the computation time in half
        estimate_units_parameters.block_sorting_parameters.skip_alignment = True

    estimate_units_parameters.check_valid(
        M=M,
        N=N,
        sampling_frequency=sampling_frequency,
        channel_locations=channel_locations,
    )

    block_size = int(
        estimate_units_parameters.block_duration_sec * sampling_frequency
    )  # size of chunks in samples
    blocks: List[TimeChunk] = get_time_chunks(
        np.int64(recording.get_num_samples()),
        chunk_size=np.int32(block_size),
        padding=np.int32(1000),
        max_num_blocks=estimate_units_parameters.max_num_blocks,
    )

    for i, block in enumerate(blocks):
        print("")
        print("=============================================")
        print(f"Processing block {i + 1} of {len(blocks)}...")
        subrecording = get_block_recording_for_scheme3(
            recording=recording,
            start_frame=int(block.start) - int(block.padding_left),
            end_frame=int(block.end) + int(block.padding_right),
        )
        subrecording_filtered = spre.bandpass_filter(
            subrecording, freq_min=300, freq_max=6000, dtype=np.float32
        )
        subrecording_preprocessed = spre.whiten(subrecording_filtered)
        aa = sorting_scheme1(
            subrecording_preprocessed,
            sorting_parameters=estimate_units_parameters.block_sorting_parameters,
            return_extra_output=True,
        )
        assert isinstance(aa, tuple)
        subsorting, extra_output = aa
        assert isinstance(subsorting, si.NumpySorting)
        labels0 = extra_output.labels
        templates0 = extra_output.templates
        peak_channel_indices0 = extra_output.peak_channel_indices
        estimate_units_block = EstimateUnitsBlock(
            start_time_sec=float(block.start / sampling_frequency),
            end_time_sec=float(block.end / sampling_frequency),
            units=[],
        )
        output.blocks.append(estimate_units_block)
        for j in range(1, np.max(labels0) + 1):
            inds0 = np.where(labels0 == j)[0]
            template0 = templates0[j - 1, :, :]
            peak_channel_ind0 = peak_channel_indices0[j - 1]
            snr0 = np.max(np.abs(template0))
            if len(inds0) > 0:
                estimate_units_block.units.append(
                    EstimateUnitsUnit(
                        unit_id=j,
                        num_spikes=len(inds0),
                        peak_channel_id=recording.get_channel_ids()[peak_channel_ind0],
                        snr=snr0
                    )
                )

    return output


def _auto_detect_channel_neighborhood_radius(
    channel_locations: np.ndarray, avg_num_channels_per_neighborhood: int
):
    M = channel_locations.shape[0]
    dist_matrix = np.zeros((M, M))
    for i in range(M):
        for j in range(M):
            dist_matrix[i, j] = np.linalg.norm(
                channel_locations[i, :] - channel_locations[j, :]
            )
    unique_distances = np.sort(np.unique(dist_matrix))
    avg_nbhd_sizes = []
    for i in range(len(unique_distances)):
        avg_nbhd_sizes.append(
            np.mean(np.sum(dist_matrix <= unique_distances[i], axis=1))
        )
    avg_nbhd_sizes = np.array(avg_nbhd_sizes)
    if np.min(avg_nbhd_sizes) > avg_num_channels_per_neighborhood:
        return float(1)
    if np.max(avg_nbhd_sizes) < avg_num_channels_per_neighborhood:
        return float(np.max(unique_distances))
    ind = np.where(avg_nbhd_sizes >= avg_num_channels_per_neighborhood)[0][0]
    print(
        "------------------- a",
        unique_distances,
        avg_nbhd_sizes,
        ind,
        unique_distances[ind],
    )
    return float(unique_distances[ind])



import numpy as np
import numpy.typing as npt
from sklearn import decomposition


def compute_pca_features(X: npt.NDArray[np.float32], *, npca: int):
    """Compute PCA features.

    Parameters
    ----------
    X : npt.NDArray[np.float32]
        Input data. L x D matrix, where L is the number of samples and D is the
        number of features.
    npca : int
        Number of PCA features to return.

    Returns
    -------
    npt.NDArray[np.float32]
        PCA features. L x npca_2 matrix where npca_2 is min(npca, L, D).
    """
    L = X.shape[0]
    D = X.shape[1]
    npca_2 = np.minimum(np.minimum(npca, L), D)
    if L == 0 or D == 0:
        return np.zeros((0, npca_2), dtype=np.float32)
    pca = decomposition.PCA(n_components=npca_2)
    return pca.fit_transform(X)



import time
from tempfile import TemporaryDirectory
import numpy as np
import spikeinterface as si
import spikeinterface.extractors as se
import spikeinterface.preprocessing as spre
import mountainsort5 as ms5
from mountainsort5.util import create_cached_recording


def test_scheme3():
    recording, sorting_true = se.toy_example(
        duration=20,
        num_channels=8,
        num_units=16,
        sampling_frequency=30000,
        num_segments=1,
        seed=0
    )  # type: ignore
    recording: si.BaseRecording = recording
    sorting_true: si.BaseSorting = sorting_true

    timer = time.time()

    # lazy preprocessing
    recording_filtered = spre.bandpass_filter(recording, freq_min=300, freq_max=6000, dtype=np.float32)
    recording_preprocessed: si.BaseRecording = spre.whiten(recording_filtered)

    with TemporaryDirectory() as tmpdir:
        # cache the recording to a temporary directory for efficient reading
        recording_cached = create_cached_recording(recording_preprocessed, folder=tmpdir)

        # sorting
        print('Starting MountainSort5')
        sorting = ms5.sorting_scheme3( # noqa
            recording_cached,
            sorting_parameters=ms5.Scheme3SortingParameters(
                block_sorting_parameters=ms5.Scheme2SortingParameters(
                    phase1_detect_channel_radius=150,
                    detect_channel_radius=50
                ),
                block_duration_sec=3
            )
        )

    elapsed_sec = time.time() - timer
    duration_sec = recording.get_total_duration()
    print(f'Elapsed time for sorting: {elapsed_sec:.2f} sec -- x{(duration_sec / elapsed_sec):.2f} speed compared with real time for {recording.get_num_channels()} channels')

    # commenting out because this step takes a while
    # print('Comparing with truth')
    # comparison = sc.compare_sorter_to_ground_truth(gt_sorting=sorting_true, tested_sorting=sorting)
    # print(comparison.get_performance())



import os
import spikeinterface as si


def create_cached_recording(recording: si.BaseRecording, *, folder: str, n_jobs: int = 1) -> si.BaseRecording:
    if not os.path.exists(folder):
        os.mkdir(folder)
    fname = f'{folder}/recording.dat'
    if recording.get_num_segments() != 1:
        raise NotImplementedError("Can only write recordings with a single segment")

    si.BinaryRecordingExtractor.write_recording(
        recording=recording,
        file_paths=[fname],
        dtype='float32',
        n_jobs=n_jobs,
        chunk_duration='5s', # not sure how to best set this
    )
    ret = si.BinaryRecordingExtractor(
        file_paths=[fname],
        sampling_frequency=recording.get_sampling_frequency(),
        channel_ids=recording.get_channel_ids(),
        num_channels=recording.get_num_channels(),
        dtype='float32'
    )
    ret.set_channel_locations(recording.get_channel_locations())
    return ret



from typing import List
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
import math
import spikeinterface as si
from .Scheme1SortingParameters import Scheme1SortingParameters
from ..core.detect_spikes import detect_spikes
from ..core.extract_snippets import extract_snippets
from ..core.isosplit6_subdivision_method import isosplit6_subdivision_method
from ..core.compute_templates import compute_templates
from ..core.compute_pca_features import compute_pca_features
from ..core.Timer import Timer


@dataclass
class SortingSchemeExtraOutput:
    templates: npt.NDArray[np.float32]  # K x T x M
    peak_channel_indices: List[int]
    times: npt.NDArray[np.int32]
    labels: npt.NDArray[np.int32]

def sorting_scheme1(
    recording: si.BaseRecording, *,
    sorting_parameters: Scheme1SortingParameters,
    return_extra_output: bool = False
):
    """MountainSort 5 sorting scheme 1

    Args:
        recording (si.BaseRecording): SpikeInterface recording object
        sorting_parameters (Scheme2SortingParameters): Sorting parameters

    Returns:
        si.BaseSorting: SpikeInterface sorting object
    """

    ###################################################################
    # Handle multi-segment recordings
    if recording.get_num_segments() > 1:
        print('Recording has multiple segments. Joining segments for sorting...')
        recording_joined = si.concatenate_recordings(recording_list=[recording])
        sorting_joined = sorting_scheme1(recording_joined, sorting_parameters=sorting_parameters)
        print('Splitting sorting into segments to match original multisegment recording...')
        sorting = si.split_sorting(sorting_joined, recording_joined)
        return sorting
    ###################################################################

    M = recording.get_num_channels()
    N = recording.get_num_frames()
    sampling_frequency = recording.sampling_frequency

    channel_locations = recording.get_channel_locations()

    print(f'Number of channels: {M}')
    print(f'Number of timepoints: {N}')
    print(f'Sampling frequency: {sampling_frequency} Hz')
    for m in range(M):
        print(f'Channel {m}: {channel_locations[m]}')

    sorting_parameters.check_valid(M=M, N=N, sampling_frequency=sampling_frequency, channel_locations=channel_locations)

    print('Loading traces')
    tt = Timer('load_traces')
    traces: np.ndarray = recording.get_traces()
    tt.report()

    print('Detecting spikes')
    tt = Timer('detect_spikes')
    time_radius = int(math.ceil(sorting_parameters.detect_time_radius_msec / 1000 * sampling_frequency))
    times, channel_indices = detect_spikes(
        traces=traces,
        channel_locations=channel_locations,
        time_radius=time_radius,
        channel_radius=sorting_parameters.detect_channel_radius,
        detect_threshold=sorting_parameters.detect_threshold,
        detect_sign=sorting_parameters.detect_sign,
        margin_left=sorting_parameters.snippet_T1,
        margin_right=sorting_parameters.snippet_T2,
        verbose=True
    )
    print(f'Detected {len(times)} spikes')
    tt.report()

    print('Removing duplicate times')
    tt = Timer('remove_duplicate_times')
    # this is important because isosplit does not do well with duplicate points
    times, channel_indices = remove_duplicate_times(times, channel_indices)
    tt.report()

    print(f'Extracting {len(times)} snippets')
    tt = Timer('extract_snippets')
    snippets = extract_snippets( # L x T x M
        traces=traces,
        channel_locations=channel_locations,
        mask_radius=sorting_parameters.snippet_mask_radius,
        times=times,
        channel_indices=channel_indices,
        T1=sorting_parameters.snippet_T1,
        T2=sorting_parameters.snippet_T2
    )
    tt.report()
    L = snippets.shape[0]
    T = snippets.shape[1]
    assert snippets.shape[2] == M

    npca = sorting_parameters.npca_per_channel * M
    print(f'Computing PCA features with npca={npca}')
    tt = Timer('compute_pca_features')
    features = compute_pca_features(snippets.reshape((L, T * M)), npca=npca)
    tt.report()

    print(f'Isosplit6 clustering with npca_per_subdivision={sorting_parameters.npca_per_subdivision}')
    tt = Timer('isosplit6_subdivision_method')
    labels = isosplit6_subdivision_method(
        X=features,
        npca_per_subdivision=sorting_parameters.npca_per_subdivision
    )
    if len(labels) > 0:
        K = int(np.max(labels))
    else:
        K = 0
    print(f'Found {K} clusters')
    tt.report()

    print('Computing templates')
    tt = Timer('compute_templates')
    templates = compute_templates(snippets=snippets, labels=labels) # K x T x M
    peak_channel_indices = [int(np.argmin(np.min(templates[i], axis=0))) for i in range(K)]
    tt.report()

    if not sorting_parameters.skip_alignment:
        print('Determining optimal alignment of templates')
        tt = Timer('align_templates')
        offsets = align_templates(templates)
        tt.report()

        print('Aligning snippets')
        tt = Timer('align_snippets')
        snippets = align_snippets(snippets, offsets, labels)
        # this is tricky - we need to subtract the offset to correspond to shifting the template
        times = offset_times(times, -offsets, labels)
        tt.report()

        print('Clustering aligned snippets')
        npca = sorting_parameters.npca_per_channel * M

        print(f'Computing PCA features with npca={npca}')
        tt = Timer('compute_pca_features')
        features = compute_pca_features(snippets.reshape((L, T * M)), npca=npca)
        tt.report()

        print(f'Isosplit6 clustering with npca_per_subdivision={sorting_parameters.npca_per_subdivision}')
        tt = Timer('isosplit6_subdivision_method')
        labels = isosplit6_subdivision_method(
            X=features,
            npca_per_subdivision=sorting_parameters.npca_per_subdivision
        )
        if len(labels) > 0:
            K = int(np.max(labels))
        else:
            K = 0
        tt.report()
        print(f'Found {K} clusters after alignment')

        print('Computing templates')
        tt = Timer('compute_templates')
        templates = compute_templates(snippets=snippets, labels=labels) # K x T x M
        peak_channel_indices = [int(np.argmin(np.min(templates[i], axis=0))) for i in range(K)]
        tt.report()

        print('Offsetting times to peak')
        tt = Timer('determine_offsets_to_peak')
        # Now we need to offset the times again so that the spike times correspond to actual peaks
        offsets_to_peak = determine_offsets_to_peak(templates, detect_sign=sorting_parameters.detect_sign, T1=sorting_parameters.snippet_T1)
        print('Offsets to peak:', offsets_to_peak)
        # This time we need to add the offset
        times = offset_times(times, offsets_to_peak, labels)
        tt.report()

    # Now we need to make sure the times are in order, because we have offset them
    print('Sorting times')
    tt = Timer('sorting times')
    sort_inds = np.argsort(times)
    times = times[sort_inds]
    labels = labels[sort_inds]
    tt.report()

    # also make sure none of the times are out of bounds now that we have offset them a couple times
    print('Removing out of bounds times')
    tt = Timer('removing out of bounds times')
    inds_okay = np.where((times >= sorting_parameters.snippet_T1) & (times < N - sorting_parameters.snippet_T2))[0]
    times = times[inds_okay]
    labels = labels[inds_okay]
    tt.report()

    print('Reordering units')
    # relabel so that units are ordered by channel
    # and we also put any labels that are not used at the end
    tt = Timer('reordering units')
    aa = np.array([float(x) for x in peak_channel_indices])
    for k in range(1, K + 1):
        inds = np.where(labels == k)[0]
        if len(inds) == 0:
            aa[k - 1] = np.Inf
    new_labels_mapping = np.argsort(np.argsort(aa)) + 1 # too tricky! my head aches right now
    labels = new_labels_mapping[labels - 1]
    tt.report()

    print('Creating sorting object')
    tt = Timer('creating sorting object')
    sorting = si.NumpySorting.from_times_labels(times_list=[times], labels_list=[labels], sampling_frequency=sampling_frequency)
    tt.report()

    if return_extra_output:
        extra_output = SortingSchemeExtraOutput(
            templates=templates,
            peak_channel_indices=peak_channel_indices,
            times=times,
            labels=labels
        )
        return sorting, extra_output
    else:
        return sorting

def remove_duplicate_times(times: npt.NDArray[np.int32], labels: npt.NDArray[np.int32]):
    if len(times) == 0:
        return times, labels
    inds = np.where(np.diff(times) > 0)[0]
    inds = np.concatenate([np.array([0]), inds + 1])
    times2 = times[inds]
    labels2 = labels[inds]
    return times2, labels2

def align_templates(templates: npt.NDArray[np.float32]):
    K = templates.shape[0]
    # T = templates.shape[1]
    # M = templates.shape[2]
    offsets = np.zeros((K,), dtype=np.int32)
    pairwise_optimal_offsets = np.zeros((K, K), dtype=np.int32)
    pairwise_inner_products = np.zeros((K, K), dtype=np.float32)
    for k1 in range(K):
        for k2 in range(K):
            offset, inner_product = compute_pairwise_optimal_offset(templates[k1], templates[k2])
            pairwise_optimal_offsets[k1, k2] = offset
            pairwise_inner_products[k1, k2] = inner_product
    for passnum in range(20):
        something_changed = False
        for k1 in range(K):
            weighted_sum = 0
            total_weight = 0
            for k2 in range(K):
                if k1 != k2:
                    offset = pairwise_optimal_offsets[k1, k2] + offsets[k2]
                    weight = pairwise_inner_products[k1, k2]
                    weighted_sum += weight * offset
                    total_weight += weight
            if total_weight > 0:
                avg_offset = int(weighted_sum / total_weight)
            else:
                avg_offset = 0
            if avg_offset != offsets[k1]:
                something_changed = True
                offsets[k1] = avg_offset
        if not something_changed:
            print('Template alignment converged.')
            break
    print('Align templates offsets: ', offsets)
    return offsets


def compute_pairwise_optimal_offset(template1: npt.NDArray[np.float32], template2: npt.NDArray[np.float32]):
    T = template1.shape[0]
    best_inner_product = -np.Inf
    best_offset = 0
    for offset in range(T):
        inner_product = np.sum(np.roll(template1, shift=offset, axis=0) * template2)
        if inner_product > best_inner_product:
            best_inner_product = inner_product
            best_offset = offset
    if best_offset > T // 2:
        best_offset = best_offset - T
    return best_offset, best_inner_product

def align_snippets(snippets: npt.NDArray[np.float32], offsets: npt.NDArray[np.int32], labels: npt.NDArray[np.int32]):
    if len(labels) == 0:
        return snippets
    snippets2 = np.zeros_like(snippets)
    for k in range(1, int(np.max(labels)) + 1):
        inds = np.where(labels == k)[0]
        snippets2[inds] = np.roll(snippets[inds], shift=offsets[k - 1], axis=1)
    return snippets2

def offset_times(times: npt.NDArray[np.int32], offsets: npt.NDArray[np.int32], labels: npt.NDArray[np.int32]):
    if len(labels) == 0:
        return times
    times2 = np.zeros_like(times)
    for k in range(1, int(np.max(labels)) + 1):
        inds = np.where(labels == k)[0]
        times2[inds] = times[inds] + offsets[k - 1]
    return times2

def determine_offsets_to_peak(templates: npt.NDArray[np.float32], *, detect_sign: int, T1: int):
    K = templates.shape[0]

    if detect_sign < 0:
        A = -templates
    elif detect_sign > 0: # pragma: no cover
        A = templates # pragma: no cover
    else:
        A = np.abs(templates) # pragma: no cover

    offsets_to_peak = np.zeros((K,), dtype=np.int32)
    for k in range(K):
        peak_channel = np.argmax(np.max(A[k], axis=0))
        peak_time = np.argmax(A[k][:, peak_channel])
        offset_to_peak = peak_time - T1
        offsets_to_peak[k] = offset_to_peak
    return offsets_to_peak



import numpy as np
import numpy.typing as npt


def compute_templates(snippets: npt.NDArray[np.float32], labels: npt.NDArray[np.int32]):
    """Compute templates from snippets and labels

    Args:
        snippets: L x T x M where L is the number of snippets, T is the number of timepoints, and M is the number of channels
        labels: array of length L with labels for the snippets

    Returns:
        templates: K x T x M where K is the number of clusters
    """
    L = snippets.shape[0]
    T = snippets.shape[1]
    M = snippets.shape[2]
    if len(labels) != L:
        raise Exception('Length of labels must equal number of snippets')
    if L == 0:
        return np.zeros((0, T, M), dtype=np.float32)
    K = int(np.max(labels))
    templates = np.zeros((K, T, M), dtype=np.float32)
    for k in range(1, K + 1):
        snippets1 = snippets[labels == k]
        templates[k - 1] = np.median(snippets1, axis=0)
    return templates



from typing import Dict, Tuple, List, Union
import numpy as np
import numpy.typing as npt
import math
import spikeinterface as si
from .Scheme2SortingParameters import Scheme2SortingParameters
from .Scheme1SortingParameters import Scheme1SortingParameters
from ..core.detect_spikes import detect_spikes
from ..core.extract_snippets import extract_snippets, extract_snippets_in_channel_neighborhood
from .sorting_scheme1 import sorting_scheme1
from ..core.SnippetClassifier import SnippetClassifier
from ..core.remove_duplicate_events import remove_duplicate_events
from ..core.get_sampled_recording_for_training import get_sampled_recording_for_training
from ..core.get_times_labels_from_sorting import get_times_labels_from_sorting
from ..core.Timer import Timer


def sorting_scheme2(
    recording: si.BaseRecording, *,
    sorting_parameters: Scheme2SortingParameters,
    return_snippet_classifiers: bool = False, # used in scheme 3
    reference_snippet_classifiers: Union[Dict[int, SnippetClassifier], None] = None, # used in scheme 3
    label_offset: int = 0 # used in scheme 3
) -> Union[si.BaseSorting, Tuple[si.BaseSorting, Dict[int, SnippetClassifier]]]:
    """MountainSort 5 sorting scheme 2

    Args:
        recording (si.BaseRecording): SpikeInterface recording object
        sorting_parameters (Scheme2SortingParameters): Sorting parameters
        return_snippet_classifiers (bool): whether to return the snippet classifiers (used in scheme 3)
        reference_snippet_classifiers: used in scheme 3
        label_offset: used in scheme 3

    Returns:
        si.BaseSorting: SpikeInterface sorting object
            or, if return_snippet_classifiers is True:
        si.BaseSorting, snippet_classifiers
    """

    ###################################################################
    # Handle multi-segment recordings
    if recording.get_num_segments() > 1:
        print('Recording has multiple segments. Joining segments for sorting...')
        recording_joined: si.BaseRecording = si.concatenate_recordings(recording_list=[recording])
        result = sorting_scheme2(
            recording_joined,
            sorting_parameters=sorting_parameters,
            return_snippet_classifiers=True,
            reference_snippet_classifiers=reference_snippet_classifiers,
            label_offset=label_offset
        )
        assert isinstance(result, tuple)
        sorting_joined, snippet_classifiers = result
        print('Splitting sorting into segments to match original multisegment recording...')
        sorting = si.split_sorting(sorting_joined, recording_joined)
        if return_snippet_classifiers:
            return sorting, snippet_classifiers
        else:
            return sorting
    ###################################################################

    M = recording.get_num_channels()
    N = recording.get_num_frames()
    sampling_frequency = recording.sampling_frequency
    channel_locations = recording.get_channel_locations()

    # check that the sorting parameters are valid
    sorting_parameters.check_valid(M=M, N=N, sampling_frequency=sampling_frequency, channel_locations=channel_locations)

    # Subsample the recording for training
    if sorting_parameters.training_duration_sec is not None:
        print(f'Using training recording of duration {sorting_parameters.training_duration_sec} sec with the sampling mode {sorting_parameters.training_recording_sampling_mode}')
        tt = Timer('SCHEME2 get_sampled_recording_for_training')
        training_recording = get_sampled_recording_for_training(
            recording=recording,
            training_duration_sec=sorting_parameters.training_duration_sec,
            mode=sorting_parameters.training_recording_sampling_mode
        )
        tt.report()
    else:
        print(f'Using the full recording for training: {recording.get_num_frames() / recording.get_sampling_frequency()} sec')
        training_recording = recording

    # Run the first phase of spike sorting (same as sorting_scheme1)
    print('Running phase 1 sorting')
    tt = Timer('SCHEME2 sorting_scheme1')
    sorting1 = sorting_scheme1(
        recording=training_recording,
        sorting_parameters=Scheme1SortingParameters(
            detect_threshold=sorting_parameters.phase1_detect_threshold,
            detect_sign=sorting_parameters.detect_sign,
            detect_time_radius_msec=sorting_parameters.phase1_detect_time_radius_msec,
            detect_channel_radius=sorting_parameters.phase1_detect_channel_radius,
            snippet_mask_radius=sorting_parameters.snippet_mask_radius,
            snippet_T1=sorting_parameters.snippet_T1,
            snippet_T2=sorting_parameters.snippet_T2,
            npca_per_channel=sorting_parameters.phase1_npca_per_channel,
            npca_per_subdivision=sorting_parameters.phase1_npca_per_subdivision
        )
    )
    assert isinstance(sorting1, si.BaseSorting)
    tt.report()

    # Get the times and labels from the first phase sorting
    tt = Timer('SCHEME2 get_times_labels_from_sorting')
    times, labels = get_times_labels_from_sorting(sorting1)
    times: np.ndarray = times
    labels: np.ndarray = labels
    if len(labels) > 0:
        K = np.max(labels) # number of clusters
        labels = labels + label_offset # used in scheme 3
    else:
        K = 0
    tt.report()

    print('Loading training traces')
    # Load the traces from the training recording
    tt = Timer('SCHEME2 training_recording.get_traces')
    training_traces: np.ndarray = training_recording.get_traces()
    training_snippets = extract_snippets(
        traces=training_traces,
        channel_locations=None,
        mask_radius=None,
        times=times,
        channel_indices=None,
        T1=sorting_parameters.snippet_T1,
        T2=sorting_parameters.snippet_T2
    )
    tt.report()

    print('Training classifier')
    # Train the classifier based on the labels obtained from the first phase sorting
    tt = Timer('SCHEME2 training classifier step 1')
    channel_masks: Dict[int, Union[List[int], None]] = {} # by channel
    for m in range(M):
        if sorting_parameters.snippet_mask_radius is not None:
            x: List[int] = []
            channel_masks[m] = []
            for m2 in range(M):
                if np.linalg.norm(channel_locations[m] - channel_locations[m2]) <= sorting_parameters.snippet_mask_radius:
                    x.append(m2)
            channel_masks[m] = x
        else:
            channel_masks[m] = None
    snippet_classifiers: Dict[int, SnippetClassifier] = {} # by channel
    for m in range(M):
        snippet_classifiers[m] = SnippetClassifier(npca=sorting_parameters.classifier_npca)
        # Add random snippets to classifier with label 0 (a noise cluster for classification)
        uniformly_spread_times = np.floor(np.linspace(sorting_parameters.snippet_T1, training_traces.shape[0] - sorting_parameters.snippet_T2 - 1, sorting_parameters.max_num_snippets_per_training_batch)).astype(np.int32)
        random_snippets = extract_snippets_in_channel_neighborhood(
            traces=training_traces,
            times=uniformly_spread_times,
            neighborhood=channel_masks[m],
            T1=sorting_parameters.snippet_T1,
            T2=sorting_parameters.snippet_T2
        )
        snippet_classifiers[m].add_training_snippets(random_snippets, label=0, offset=0)
    tt.report()

    # Add the snippets from the first phase sorting to the classifier
    print('Adding snippets from phase 1 sorting')
    tt = Timer('SCHEME2 adding snippets from phase 1 sorting')
    for k in range(label_offset + 1, label_offset + K + 1):
        inds0 = np.where(labels == k)[0]
        snippets0 = training_snippets[inds0]
        template0 = np.median(snippets0, axis=0) # T x M
        if sorting_parameters.detect_sign < 0:
            AA = -template0
        elif sorting_parameters.detect_sign > 0: # pragma: no cover
            AA = template0 # pragma: no cover
        else:
            AA = np.abs(template0) # pragma: no cover
        peak_indices_over_channels = np.argmax(AA, axis=0)
        peak_values_over_channels = np.max(AA, axis=0)
        peaks_to_include: List[dict] = []
        for m in range(M):
            if peak_values_over_channels[m] >= sorting_parameters.detect_threshold * 0.4: # should be a parameter
                peaks_to_include.append({
                    'channel': m,
                    'offset': peak_indices_over_channels[m] - sorting_parameters.snippet_T1
                })
        for peak in peaks_to_include:
            m = peak['channel']
            offset = peak['offset']
            if channel_masks[m] is not None:
                snippets0_masked = snippets0[:, :, channel_masks[m]]
            else:
                snippets0_masked = snippets0
            snippet_classifiers[m].add_training_snippets(
                snippets=subsample_snippets(np.roll(snippets0_masked, shift=-offset, axis=1), sorting_parameters.max_num_snippets_per_training_batch),
                label=k,
                offset=offset
            )
    training_snippets = None # Free up memory

    print('Fitting models')
    tt = Timer('SCHEME2 fitting models')
    for m in range(M):
        snippet_classifiers[m].fit()
    tt.report()

    # Now that we have the classifier, we can do the full sorting
    # Iterate over time chunks, detect and classify all spikes, and collect the results

    if sorting_parameters.classification_chunk_sec is None:
        chunk_size = int(math.ceil(100e6 / recording.get_num_channels())) # size of chunks in samples
    else:
        chunk_size = int(math.ceil(sorting_parameters.classification_chunk_sec * recording.sampling_frequency))

    print(f'Chunk size: {chunk_size / recording.sampling_frequency} sec')
    chunks = get_time_chunks(np.int64(recording.get_num_samples()), chunk_size=np.int32(chunk_size), padding=np.int32(1000))
    times_list: list[npt.NDArray[np.int64]] = []
    labels_list: list[npt.NDArray[np.int32]] = []
    labels_reference_list = [] if reference_snippet_classifiers is not None else None
    for i, chunk in enumerate(chunks):
        print(f'Time chunk {i + 1} of {len(chunks)}')
        print('Loading traces')
        tt = Timer('SCHEME2 loading traces')
        traces_chunk: np.ndarray = recording.get_traces(start_frame=int(chunk.start - chunk.padding_left), end_frame=int(chunk.end + chunk.padding_right))
        tt.report()

        print('Detecting spikes')
        tt = Timer('SCHEME2 detecting spikes')
        time_radius = int(math.ceil(sorting_parameters.detect_time_radius_msec / 1000 * sampling_frequency))
        times_chunk, channel_indices_chunk = detect_spikes(
            traces=traces_chunk,
            channel_locations=channel_locations,
            time_radius=time_radius,
            channel_radius=sorting_parameters.detect_channel_radius,
            detect_threshold=sorting_parameters.detect_threshold,
            detect_sign=sorting_parameters.detect_sign,
            margin_left=sorting_parameters.snippet_T1,
            margin_right=sorting_parameters.snippet_T2,
            verbose=False
        )
        print(f'Scheme 2 detected {len(times_chunk)} spikes in chunk {i + 1} of {len(chunks)}')
        tt.report()

        print('Extracting and classifying snippets')
        tt = Timer('SCHEME2 extracting and classifying snippets')
        labels_chunk = np.zeros(len(times_chunk), dtype='int32')
        labels_reference_chunk = np.zeros(len(times_chunk), dtype='int32') if reference_snippet_classifiers is not None else None
        for m in range(M):
            inds = np.where(channel_indices_chunk == m)[0]
            if len(inds) > 0:
                snippets2 = extract_snippets_in_channel_neighborhood(
                    traces=traces_chunk,
                    times=times_chunk[inds],
                    neighborhood=channel_masks[m],
                    T1=sorting_parameters.snippet_T1,
                    T2=sorting_parameters.snippet_T2
                )
                labels_chunk_m, offsets_chunk_m = snippet_classifiers[m].classify_snippets(snippets2)
                if labels_chunk_m is not None and offsets_chunk_m is not None:
                    labels_chunk[inds] = labels_chunk_m
                    times_chunk[inds] = times_chunk[inds] - offsets_chunk_m
                if reference_snippet_classifiers is not None:
                    assert labels_reference_chunk is not None
                    labels_reference_chunk_m, _ = reference_snippet_classifiers[m].classify_snippets(snippets2)
                    if labels_reference_chunk_m is not None:
                        labels_reference_chunk[inds] = labels_reference_chunk_m
        tt.report()

        print('Updating events')
        tt = Timer('SCHEME2 updating events')

        # remove events with label 0
        valid_inds = np.where(labels_chunk > 0)[0]
        times_chunk: npt.NDArray[np.int32] = times_chunk[valid_inds]
        labels_chunk: npt.NDArray[np.int32] = labels_chunk[valid_inds]
        labels_reference_chunk = labels_reference_chunk[valid_inds] if labels_reference_chunk is not None else None

        # now that we offset them we need to re-sort
        sort_inds2 = np.argsort(times_chunk)
        times_chunk: npt.NDArray[np.int32] = times_chunk[sort_inds2]
        labels_chunk: npt.NDArray[np.int32] = labels_chunk[sort_inds2]
        labels_reference_chunk = labels_reference_chunk[sort_inds2] if labels_reference_chunk is not None else None

        print('Removing duplicates')
        new_inds = remove_duplicate_events(times_chunk, labels_chunk, tol=time_radius)
        times_chunk: npt.NDArray[np.int32] = times_chunk[new_inds]
        labels_chunk: npt.NDArray[np.int32] = labels_chunk[new_inds]
        labels_reference_chunk = labels_reference_chunk[new_inds] if labels_reference_chunk is not None else None

        # remove events in the margins
        valid_inds = np.where((chunk.padding_left <= times_chunk) & (times_chunk < chunk.total_size - chunk.padding_right))[0]
        times_chunk: npt.NDArray[np.int32] = times_chunk[valid_inds]
        labels_chunk: npt.NDArray[np.int32] = labels_chunk[valid_inds]
        labels_reference_chunk = labels_reference_chunk[valid_inds] if labels_reference_chunk is not None else None

        # don't forget to cast to int64 add the chunk start time
        times_list.append(
            times_chunk.astype(np.int64) + chunk.start - chunk.padding_left
        )
        labels_list.append(labels_chunk)
        if reference_snippet_classifiers is not None:
            assert labels_reference_list is not None
            labels_reference_list.append(labels_reference_chunk)

        tt.report()

    # Now concatenate the results
    print('Concatenating results')
    tt = Timer('SCHEME2 concatenating results')
    times_concat: npt.NDArray[np.int64] = np.concatenate(times_list)
    labels_concat: npt.NDArray[np.int32] = np.concatenate(labels_list)
    labels_reference_concat = np.concatenate(labels_reference_list) if labels_reference_list is not None else None
    tt.report()

    print('Perorming label mapping')
    tt = Timer('SCHEME2 label mapping')
    if reference_snippet_classifiers is not None:
        assert labels_reference_concat is not None
        mapping = get_labels_to_reference_labels_mapping(labels_concat, labels_reference_concat, label_offset=label_offset)
        print('==== mapping =======================')
        for k1, k2 in mapping.items():
            print(f'{k1} -> {k2}')
        print('====================================')
        for m in range(M):
            snippet_classifiers[m].apply_label_mapping(mapping)
        for k1, k2 in mapping.items():
            labels_concat[labels_concat == k1] = k2
    tt.report()

    # Now create a new sorting object from the times and labels results
    print('Creating sorting object')
    tt = Timer('SCHEME2 creating sorting object')
    sorting2 = si.NumpySorting.from_times_labels([times_concat], [labels_concat], sampling_frequency=recording.sampling_frequency)
    tt.report()

    if return_snippet_classifiers:
        return sorting2, snippet_classifiers
    else:
        return sorting2

# Here's what this function does:
# 1. For each unit, find the matching unit in the reference (has to be a MUTUAL >0.5 overlap)
# 2. If the matching unit is found, then map the unit to the matching unit
# 3. If the matching unit is not found, then map it to the smallest unused label starting with label_offset+1
def get_labels_to_reference_labels_mapping(labels: npt.NDArray[np.int32], labels_reference: npt.NDArray[np.int32], *, label_offset: int) -> Dict[int, int]:
    mapping: Dict[int, Union[int, None]] = {}
    unique_labels = np.sort(np.unique(labels))
    last_used_k = label_offset
    for k in unique_labels:
        mapping[k] = None # initialize to None, if it stays as None, then we will need to create a new label
        inds = np.where(labels == k)[0]
        a = labels_reference[inds]
        k_refs, k_ref_counts = np.unique(a, return_counts=True)
        for ii in range(len(k_refs)):
            if k_ref_counts[ii] > 0.5 * len(inds): # the 0.5 is chosen so we don't map to the same unit twice
                inds_ref = np.where(labels_reference == k_refs[ii])[0] # import to test the other way around
                if k_ref_counts[ii] > 0.5 * len(inds_ref): # mutual overlap
                    mapping[k] = k_refs[ii] # map to the reference label
                    break
        if mapping[k] is None: # if not mapped to reference label, then create a new label
            mapping[k] = last_used_k + 1
            last_used_k = last_used_k + 1

    # do this for typing reasons
    mapping2: Dict[int, int] = {}
    for k, k2 in mapping.items():
        assert k2 is not None
        mapping2[k] = k2
    return mapping2

class TimeChunk:
    def __init__(self, start: np.int64, end: np.int64, padding_left: np.int32, padding_right: np.int32):
        self.start = start
        self.end = end
        self.padding_left = padding_left
        self.padding_right = padding_right
        self.total_size = self.end - self.start + np.int64(padding_left) + np.int64(padding_right)

def get_time_chunks(num_samples: np.int64, chunk_size: np.int32, padding: np.int32, max_num_blocks: Union[int, None] = None) -> List[TimeChunk]:
    """Get time chunks
    Inputs:
        num_samples: number of samples in the recording
        chunk_size: size of each chunk in samples
        padding: padding on each side of the chunk in samples
        max_num_blocks: maximum number of blocks to return or None for no limit - if not None, spacing between blocks will be added
    Returns:
        chunks: list of TimeChunk objects
    """
    spacing = 0
    if max_num_blocks is not None:
        if chunk_size * max_num_blocks < num_samples:
            spacing = np.int64(np.floor((num_samples - chunk_size * max_num_blocks) / (max_num_blocks - 1)))
    chunks = []
    start = np.int64(0)
    while start < num_samples:
        end = np.int64(start) + np.int64(chunk_size)
        if end > num_samples:
            end = num_samples
        if end + chunk_size / 2 > num_samples:
            # if the next chunk will be too small, then just make this the last chunk
            end = num_samples
        padding_left = np.minimum(padding, start)
        padding_right = np.minimum(padding, num_samples - end)
        chunks.append(TimeChunk(start=start, end=end, padding_left=padding_left, padding_right=padding_right))
        start = end
        if spacing > 0:
            start = start + spacing
        if max_num_blocks is not None and len(chunks) >= max_num_blocks:
            break
    return chunks

def subsample_snippets(snippets: npt.NDArray[np.float32], max_num: int) -> np.ndarray:
    """Subsample snippets
    Inputs:
        snippets: 3D array of snippets (num_snippets x T x M)
        max_num: maximum number of snippets to return
    Returns:
        snippets_out: 3D array of snippets (num_snippets x T x M)
    """
    num_snippets = snippets.shape[0]
    if num_snippets > max_num:
        inds = np.arange(0, max_num) * num_snippets // max_num
        snippets_out = snippets[inds]
    else:
        snippets_out = snippets
    return snippets_out



import time


class Timer:
    def __init__(self, label: str):
        self._label = label
        self._start_time = time.time()
    def report(self):
        elapsed = time.time() - self._start_time
        print(f'*** MS5 Elapsed time for {self._label}: {elapsed:.3f} seconds ***')



import importlib.metadata
__version__ = importlib.metadata.version("mountainsort5")

from .schemes.sorting_scheme1 import sorting_scheme1 # noqa: F401
from .schemes.Scheme1SortingParameters import Scheme1SortingParameters # noqa: F401
from .schemes.sorting_scheme2 import sorting_scheme2 # noqa: F401
from .schemes.Scheme2SortingParameters import Scheme2SortingParameters # noqa: F401
from .schemes.sorting_scheme3 import sorting_scheme3 # noqa: F401
from .schemes.Scheme3SortingParameters import Scheme3SortingParameters # noqa: F401



import numpy as np
import numpy.typing as npt


def remove_duplicate_events(times: npt.NDArray[np.int32], labels: npt.NDArray[np.int32], *, tol: int) -> npt.NDArray[np.int32]:
    new_labels = np.array(labels)
    unit_ids = np.unique(new_labels)
    for unit_id in unit_ids:
        unit_inds = np.nonzero(new_labels == unit_id)[0]
        unit_times = times[unit_inds]
        inds_duplicate = find_duplicate_times(unit_times, tol=tol)
        new_labels[unit_inds[inds_duplicate]] = 0
    inds_nonzero = np.nonzero(new_labels)[0]
    return inds_nonzero

def find_duplicate_times(times: npt.NDArray[np.int32], *, tol: int) -> npt.NDArray[np.int32]:
    ret: list[np.int32] = []
    deleted = np.zeros((len(times),), dtype=np.int16)
    for i1 in range(len(times)):
        if not deleted[i1]:
            i2 = i1 + 1
            while i2 < len(times) and times[i2] <= times[i1] + tol:
                ret.append(np.int32(i2))
                deleted[i2] = True
                i2 += 1
    return np.array(ret, dtype=np.int32)




import numpy as np
import numpy.typing as npt
import warnings
from scipy.cluster.hierarchy import ClusterWarning
from scipy.spatial.distance import squareform
from isosplit6 import isosplit6
from .compute_pca_features import compute_pca_features


warnings.filterwarnings('ignore', category=ClusterWarning)


from typing import Union

def isosplit6_subdivision_method(
    X: npt.NDArray[np.float32], *,
    npca_per_subdivision: int,
    inds: Union[npt.NDArray[np.int32], None] = None # pass in inds so that we don't keep making copies of the array
):
    """Isosplit6 subdivision method

    Args:
        X: L x M array where L is the number of data points and M is the number of features
        npca_per_subdivision: number of pca components to use for each subdivision
        inds: indices of the data points to use (optional)

    Returns:
        _type_: _description_
    """
    if inds is not None:
        X_sub = X[inds]
    else:
        X_sub = X
    L = X_sub.shape[0]
    if L == 0:
        return np.zeros((0,), dtype=np.int32)
    features = compute_pca_features(X_sub, npca=npca_per_subdivision)
    labels = isosplit6(features)
    if len(labels) > 0:
        K = int(np.max(labels))
    else:
        K = 0
    if K <= 1:
        return labels
    centroids = np.zeros((K, X.shape[1]), dtype=np.float32)
    for k in range(1, K + 1):
        centroids[k - 1] = np.median(X_sub[labels == k], axis=0)
    X_sub = None # free up memory
    dists_between_centroids = np.sqrt(np.sum((centroids[:, None, :] - centroids[None, :, :]) ** 2, axis=2))

    # Before I added the following line I was getting warnings: "The symmetric non-negative hollow observation matrix looks suspiciously like an uncondensed distance matrix"
    # so I think this is needed
    # because linkage function expects "y is a 1-D condensed distance matrix,"
    dists_between_centroids = squareform(dists_between_centroids)

    # hierarchical clustering
    from scipy.cluster.hierarchy import linkage, cut_tree
    Z = linkage(dists_between_centroids, method='single', metric='euclidean')
    clusters0 = cut_tree(Z, n_clusters=2)
    cluster_inds_1 = np.where(clusters0 == 0)[0] + 1
    cluster_inds_2 = np.where(clusters0 == 1)[0] + 1
    inds1 = np.where(np.isin(labels, cluster_inds_1))[0]
    inds2 = np.where(np.isin(labels, cluster_inds_2))[0]
    if inds is not None:
        inds1_b = inds[inds1]
        inds2_b = inds[inds2]
    else:
        inds1_b = inds1
        inds2_b = inds2
    labels1 = isosplit6_subdivision_method(X, npca_per_subdivision=npca_per_subdivision, inds=inds1_b)
    labels2 = isosplit6_subdivision_method(X, npca_per_subdivision=npca_per_subdivision, inds=inds2_b)
    K1 = int(np.max(labels1))
    # K2 = int(np.max(labels2))
    ret_labels = np.zeros(L, dtype=np.int32)
    ret_labels[inds1] = labels1
    ret_labels[inds2] = labels2 + K1
    return ret_labels
